""" 
Steam Machine

"""
import unittest, random

ACCEPTABLE_ERROR = 0.01



class SlotHolder:
    def __init__(self):
        self.slot_instances = {}
        for name, kind in self.slots.items():
            self.slot_instances[ name ] = kind(self, name)
    
    def __getitem__(self, item):
        if not item in self.slots:
            raise ConnectionError("No such slot:"+str(item))
        if not item in self.slot_instances:
            raise MachineError("Slot with no slot instance:"+str(item))
        return self.slot_instances[ item ] 
      
class Machine(SlotHolder):
    
    def __init__(self, slots, debug=0):
        self.debug = debug
        self.name = "MACHINE"
        self.slots = slots.copy()
        self.slots["done"]=PipeSlot
        SlotHolder.__init__(self)
        
        self.components = set()
        self.connections = []
        self.connected_slots = set()
        
        self.validator = None
        
    def setValidator(self, validator):
        self.validator = validator
        
    def validate(self):
        if self.validator is None:
            raise MachineError("Machine does not have a validator")
        return self.validator(self)
        
    def addComponent(self, component):
        self.components.add( component )

    def connect(self, slot_a, slot_b):
        if slot_a == slot_b:
            raise ConnectionError("Cant connect a slot to itself")
        if slot_a in self.connected_slots:
            raise ConnectionError("Slot already connected: "+str(slot_a))
        if slot_b in self.connected_slots:
            raise ConnectionError("Slot already connected: "+str(slot_b))
        if slot_a.kind != slot_b.kind:
            raise ConnectionError("Cannot conect slots of different kind")
        
        self.connections.append( (slot_a, slot_b) )
        
        self.connected_slots.add( slot_a )
        self.connected_slots.add( slot_b )
        
            
    def reset(self):
        for c in self.components:
            c.reset()

    def dump(self):
        print "COMPONENTS:"
        for c in self.components:
            print "\t", c
        print "SLOTS:"
        for a,b in self.connections:
            print a,"  <--->  ", b
    def step(self, num_steps = 1):
        for step in range(num_steps):
            for slot_a, slot_b in self.connections:
                slot_a.in_value = slot_b.out_value
                slot_b.in_value = slot_a.out_value
                
            for component in self.components:
                component.step()
                
            
            if self.debug:
                print "STEP:", step, "-"*50
                print "MACHINE:"
                for s in self.slot_instances.values():
                    print "\t", s
                for c in self.components:
                    c.dump()
                
class Component(SlotHolder):
    slots = []
    broken = False
    
    def __init__(self, name=""):
        SlotHolder.__init__(self)
        self.validators = []
        self.name = name
        
        for k in dir(self):
            if isinstance(getattr(self, k), self.__init__.__class__):
                if "validate_" == k[:9]:
                    self.validators.append( getattr(self, k) )
        self.status_init()
        
    def dump(self):
        print self, ["working", "broken"][self.broken]
        for slot in self.slot_instances.values():
            print "\t", slot
        
    def reset(self):
        for s in self.slot_instances.values():
            s.reset()
        self.broken = False
        self.status_reset()
    
    def step(self):
        for valid in self.validators:
            valid()
        if self.broken:
            return
        self.status_step()
    # Method from here on must be subclassed
        
    def status_init(self):
        pass
    
    def status_reset(self):
        pass
    
    def status_step(self):
        raise NotImplementedError()
    
    def __repr__(self):
        return "<"+self.__class__.__name__+":"+ [str(id(self)), self.name][self.name!=""]+" >"
    
class ComponentOutValidatorMixin:
    # replace in subclass with slot names that brake the machine if they have input
    outputs = []
    
    def validate_outputsnoinput(self):
        for slot in self.outputs:
            if self[slot].in_value != 0:
                self.broken = True
                return False
        return True
            
class Slot:
    """" 
    each slot class must define its kind name, slots can only be connected to
    slots of the same kind
    """
    kind = "generic"
    
    def __init__(self, component, name):
        self.component = component
        self.name = name
        self.reset()
        
    def reset(self):
        self.in_value = 0
        self.out_value = 0
        
    def __str__(self):
        name = "<slot %s of %s"%(self.name, [self.component.name, "%s (%s)"%(self.component.__class__.__name__, id(self.component))][self.component.name == ""])
        return "%50s| in: %i out %i>"%(name, self.in_value, self.out_value)
    
class PipeSlot(Slot):
    kind = "pipe"
    
class GearSlot(Slot):
    kind = "gear"
    
class MachineError(Exception):
    pass

class ConnectionError(MachineError):
    pass

###
#
#
# Components go here
#
#
###
class BoilerComponent(Component, ComponentOutValidatorMixin):
    slots = dict(output=PipeSlot)
    outputs = ["output"]
    
    def status_step(self):
        self["output"].out_value = 1000
        
  
class GaugeComponent(Component, ComponentOutValidatorMixin):
    slots = dict(input=PipeSlot, output=PipeSlot)
    outputs = ["output"]
    
    def status_init(self):
        self.value = 0
        
    def status_step(self):
        self["output"].out_value = self["input"].in_value
        self.value = self["input"].in_value
        
class ElbowComponent(Component):
    slots = dict(A=PipeSlot, B=PipeSlot)
    
    def status_step(self):
        if self["A"].in_value and self["B"].in_value:
            self.broken = True
            return
        if self["A"].in_value:
            self["B"].out_value = self["A"].in_value
            
        if self["B"].in_value:
            self["A"].out_value = self["B"].in_value

class ElbowGearComponent(ElbowComponent):
    slots = dict(A=GearSlot, B=GearSlot)
    
    
class TComponent(Component):
    slots = dict(A=PipeSlot, B=PipeSlot, C=PipeSlot)
    
    def status_step(self):
        total_input = 0
        active_inputs = 0
        
        for slot in self.slots:
            if self[slot].in_value:
                total_input += self[slot].in_value
                active_inputs += 1
            
        if active_inputs == len(self.slots):
            self.broken = True
            return
        
        if not active_inputs:
            return
        
        for slot in self.slots:
            if not self[slot].in_value:
                self[slot].out_value = total_input / (len(self.slots)-active_inputs)
            else:
                self[slot].out_value = 0
            
class XComponent(TComponent):
    slots = dict(A=PipeSlot, B=PipeSlot, C=PipeSlot, D=PipeSlot)
            
    
###
      
class ManualGearComponent(Component):
    slots = dict(gear=GearSlot)
    
    def setValue(self, value):
        if value < 0: value = 0
        if value > 1000: value = 1000
        self["gear"].out_value = value
    def getValue(self):
        return self["gear"].out_value
    value = property(getValue, setValue)
    
    def status_step(self): pass
    
class GearValveComponent(Component):
    slots = dict(A=PipeSlot, B=PipeSlot, escape=PipeSlot, gear=GearSlot)
    
    def status_step(self):
        if self["A"].in_value and self["B"].in_value:
            if self["A"].in_value + self["B"].in_value > self.value:
                self.broken = True
                return
            self["escape"].out_value = self["A"].in_value + self["B"].in_value
            return
        
        if self["A"].in_value:
            self["B"].out_value = max(self["A"].in_value-self.value, 0)
            self["escape"].out_value = min( self.value, self["A"].in_value)
    
        if self["B"].in_value:
            self["A"].out_value = max(self["B"].in_value-self.value, 0)
            self["escape"].out_value = min( self.value, self["B"].in_value)
    
    def getValue(self):
        return self["gear"].in_value
    
    
    value = property(getValue)
    
        
class PressureToGearComponent(Component):
    slots = dict(input=PipeSlot, output=PipeSlot, gear=GearSlot)
    
    def status_step(self):
        if self["output"].in_value:
            self.broken = True
            return
        
        self["gear"].out_value = self["input"].in_value
        self["output"].out_value = self["input"].in_value
        
    
class TankComponent(Component):
    slots = dict(input=PipeSlot, output=GearSlot, reset=PipeSlot)
    
    capacity = 1000
    reset_value = 500
    
    def status_init(self):
        self.pressure = 0
        
    status_reset = status_init
    
    def status_step(self):
        if self["reset"].in_value > self.reset_value:
            self.pressure = 0
            return
        
        if self.pressure + self["input"].in_value > self.capacity:
            self.broken = True
            
        self.pressure += self["input"].in_value
        self["output"].out_value = self.pressure
        
class SmallTankComponent(TankComponent):
    capacity = 500
    reset_value = 500
###
#
#
# Unitests go here
#
#
###

class TestSimpleMachine(unittest.TestCase):
    
    def setUp(self):
        self.m = m = Machine({"MIL":PipeSlot})
        m.setValidator(lambda s: s["MIL"].in_value == 1000)
        
    def testSimpleSolution(self):
        m = self.m
        b = BoilerComponent()
        
        m.addComponent( b )
        m.connect( b["output"], m["MIL"] )
        m.step(20)
        
        self.assertEqual(m.validate(), True)
    
    def testBreakBoiler(self):
        m = self.m
        
        b = BoilerComponent()
        m.addComponent( b )
        
        b2 = BoilerComponent()
        m.addComponent( b2 )
        m.connect( b2["output"], b["output"] )
        
        m.step(20)
        
        self.assertEqual(b.broken, True)
     
    def testGaugedSimpleSolution(self):
        m = self.m
        b = BoilerComponent()
        g = GaugeComponent()
        m.addComponent( b )
        m.addComponent( g )
        m.connect( b["output"], g["input"] )
        m.connect( g["output"], m["MIL"] )
        m.step(20)
        
        self.assertEqual(m.validate(), True)   
        self.assertEqual(g.value, 1000)
        
    def testElbowSimpleSolution(self):
        m = self.m
        b = BoilerComponent()
        e = ElbowComponent()
        m.addComponent( b )
        m.addComponent( e )
        m.connect( b["output"], e["A"] )
        m.connect( e["B"], m["MIL"] )
        m.step(20)
        
        self.assertEqual(m.validate(), True)   
        
    def testElbowBSimpleSolution(self):
        m = self.m
        b = BoilerComponent()
        e = ElbowComponent()
        m.addComponent( b )
        m.addComponent( e )
        m.connect( b["output"], e["B"] )
        m.connect( e["A"], m["MIL"] )
        m.step(20)
        
        self.assertEqual(m.validate(), True)   
        
    def testBreakElbow(self):
        m = self.m
        
        b = BoilerComponent()
        m.addComponent( b )
        
        b2 = BoilerComponent()
        m.addComponent( b2 )
        
        e = ElbowComponent()
        m.addComponent( e )
        m.connect( b2["output"], e["A"] )
        m.connect( b["output"], e["B"] )
        
        m.step(20)
        
        self.assertEqual(b.broken, False)
        self.assertEqual(b2.broken, False)
        self.assertEqual(e.broken, True)
        
    def testErrors(self):
        m = self.m
        b = BoilerComponent()
        m.addComponent( b )
        
        b2 = BoilerComponent()
        m.addComponent( b2 )
        
        e = ElbowComponent()
        m.addComponent( e )
        
        self.assertRaises(ConnectionError, lambda: m.connect(e["Z"], e["A"]))
        m.connect(b["output"], e["A"])
        self.assertRaises(ConnectionError, lambda: m.connect(b2["output"], e["A"]))
        self.assertRaises(ConnectionError, lambda: m.connect(e["A"], e["A"]))
        
        w = ManualGearComponent()
        m.addComponent( w )
        self.assertRaises(ConnectionError, lambda: m.connect(w["gear"], m["MIL"]))
    
        
class TestSumMachine(unittest.TestCase):
    
    def setUp(self):
        self.m = m = Machine({"A":PipeSlot, "B":PipeSlot, "A+B":PipeSlot})
        m.setValidator(lambda s: s["A+B"].in_value == s["A"].out_value + s["B"].out_value)
        m["A"].out_value = random.randint(0,500)
        m["B"].out_value = random.randint(0,500)
        
    def testSolution(self):
        m = self.m
        
        t = TComponent()
        m.addComponent( t )
        
        m.connect( m["A"], t["A"])
        m.connect( m["B"], t["B"])
        m.connect( m["A+B"], t["C"])
        
        m.step(10)
        
        self.assertNotEqual(t.broken, True)
        self.assertEqual(m.validate(), True)   
        
    def testBrokenT(self):
        m = self.m
        
        t = TComponent()
        m.addComponent( t )
        
        b = BoilerComponent()
        m.addComponent( b )
        m.connect( m["A"], t["A"])
        m.connect( m["B"], t["B"])
        m.connect( b["output"], t["C"])
        
        m.step(10)
        
        self.assertEqual(t.broken, True)
        
   
        
class TestSumDiv2Machine(unittest.TestCase):
    
    def setUp(self):
        self.m = m = Machine({"A":PipeSlot, "B":PipeSlot, "A+B":PipeSlot})
        m.setValidator(lambda s: abs(s["A+B"].in_value - (s["A"].out_value + s["B"].out_value)/2) < ACCEPTABLE_ERROR)
        m["A"].out_value = random.randint(0,1000)
        m["B"].out_value = random.randint(0,1000)
        
    def testSolution(self):
        m = self.m
        
        t = XComponent()
        m.addComponent( t )
        
        m.connect( m["A"], t["A"])
        m.connect( m["B"], t["B"])
        m.connect( m["A+B"], t["C"])
        
        m.step(10)
        
        self.assertNotEqual(t.broken, True)
        self.assertEqual(m.validate(), True)   
        
    def testBrokenX(self):
        m = self.m
        
        t = XComponent()
        m.addComponent( t )
        
        b = BoilerComponent()
        m.addComponent( b )
        b2 = BoilerComponent()
        m.addComponent( b2 )
        
        m.connect( m["A"], t["A"])
        m.connect( m["B"], t["B"])
        m.connect( b["output"], t["C"])
        m.connect( b2["output"], t["D"])
        
        m.step(10)
        
        self.assertEqual(t.broken, True)

class TestGearValve(unittest.TestCase):
    
    def setUp(self):
        self.m = m = Machine({"A":PipeSlot, "B":PipeSlot, "A-30":PipeSlot})
        m.setValidator(lambda s: s["A-30"].in_value == max(s["A"].out_value-30,0))
        m["A"].out_value = random.randint(0,500)
        m["B"].out_value = random.randint(0,500)
        
   
    def testSolution(self):
        m = self.m
        
        t = GearValveComponent()
        m.addComponent( t )
        t['gear'].in_value = 30
        m.connect( m["A"], t["A"])
        m.connect( m["A-30"], t["B"])
        
        m.step(10)
        
        self.assertEqual(m.validate(), True)   
        
    def testEscape(self):
        m = self.m
        
        t = GearValveComponent()
        m.addComponent( t )
        t['gear'].in_value = 30
        m.connect( m["A"], t["A"])
        m.connect( m["A-30"], t["B"])
        m["A"].out_value = 15
        m.step(10)
        self.assertEqual(t["escape"].out_value, 15)   
        m["A"].out_value = 100
        m.step(10)
        self.assertEqual(t["escape"].out_value, 30)  
        
    def testBreak(self):
        m = self.m
        
        t = GearValveComponent()
        m.addComponent( t )
        t['gear'].in_value = 30
        m.connect( m["A"], t["A"])
        m.connect( m["B"], t["B"])
        
        m.step(10)
        
        self.assertEqual(t.broken, True)  
        
class TestPressureGear(unittest.TestCase):
    
    def setUp(self):
        self.m = m = Machine({"A":PipeSlot, "B":PipeSlot, "A-B":PipeSlot})
        m.setValidator(lambda s: s["A-B"].in_value == max(s["A"].out_value-s["B"].out_value,0))
        m["A"].out_value = random.randint(0,500)
        m["B"].out_value = random.randint(0,500)
        
    def testSolution(self):
        m = self.m
        gv = GearValveComponent()
        pg = PressureToGearComponent()
        m.addComponent(gv)
        m.addComponent(pg)
        
        m.connect(gv["gear"], pg["gear"])
        m.connect(gv["A"], m["A"])
        m.connect(pg["input"], m["B"])
        m.connect(gv["B"], m["A-B"])
        
        m.step(10)
        self.assertEqual(m.validate(), True)  
         
class DebugTests(unittest.TestCase):
    
    def setUp(self):
        self.m = m = Machine({"A":PipeSlot, "B":PipeSlot, "A-B":PipeSlot})  
        
    def testTank(self):
        m = self.m
        t = TankComponent()
        m.addComponent( t )
        m["A"].out_value = 200
        m.connect(m["A"], t["input"])   
        
        m.step(); self.assertEqual( t.pressure , 200 )
        m.step(); self.assertEqual( t.pressure , 400 )
        m.step(); self.assertEqual( t.pressure , 600 )
        m.step(); self.assertEqual( t.pressure , 800 )
        m.step(); self.assertEqual( t.pressure , 1000 )
        m.step(); self.assertEqual( t.broken , True )
        
    def testResetTank(self):
        m = self.m
        t = TankComponent()
        m.addComponent( t )
        m["A"].out_value = 200
        m.connect(m["A"], t["input"])   
        m.connect(m["B"], t["reset"])
        m.step(); self.assertEqual( t.pressure , 200 )
        m.step(); self.assertEqual( t.pressure , 400 )
        m.step(); self.assertEqual( t.pressure , 600 )
        m["A"].out_value = 0
        m.step(); self.assertEqual( t.pressure , 600 )
        m["B"].out_value = 700
        m.step(); self.assertEqual( t.pressure , 0)
        
        
class MultiplicationMachine(unittest.TestCase):
    
    def setUp(self):
        self.m = m = Machine({"A":PipeSlot, "B":PipeSlot, "A*B":PipeSlot, "done":PipeSlot}, 0)
        m.setValidator(lambda s: s["A*B"].in_value == max(s["A"].out_value*s["B"].out_value,0))
        
        v1 = 1000#10
        v2 = 1000#5
        while v1 * v2 > 1000:
            v1 = random.randint(0,1000)
            v2 = random.randint(0,1000)
        m["A"].out_value = v1
        m["B"].out_value = v2
        
        boiler1 = BoilerComponent("boiler1"); m.addComponent( boiler1 )
        boiler2 = BoilerComponent("boiler2"); m.addComponent( boiler2 )
        boiler3 = BoilerComponent("boiler2"); m.addComponent( boiler3 )
        
        tank1 = TankComponent("tank1"); m.addComponent( tank1 )
        tank2 = TankComponent("tank2"); m.addComponent( tank2 )
        
        gearvalve1 = GearValveComponent("gearvalve1"); m.addComponent( gearvalve1 )
        gearvalve2 = GearValveComponent("gearvalve2"); m.addComponent( gearvalve2 )
        gearvalve3 = GearValveComponent("gearvalve3"); m.addComponent( gearvalve3 )
        gearvalve4 = GearValveComponent("gearvalve4"); m.addComponent( gearvalve4 )
        
        manualgear = ManualGearComponent("manualgear"); m.addComponent( manualgear )
        
        pressuregear = PressureToGearComponent("pressuregear"); m.addComponent( pressuregear )

        def makeDelays(steps, name="delay"):
            delays = []
            for x in range(steps):
                c = ElbowComponent(name+"-"+str(x))
                m.addComponent(c)
                
                if delays:
                    m.connect( delays[-1]["B"], c["A"] )
                delays.append( c )
            return delays
        
        delays = makeDelays(6, "delay1")
        delays2 = makeDelays(4, "delay2")
        delays3 = makeDelays(2, "delay3")
        delays4 = makeDelays(1, "delay4")
        
        m.connect( m["B"], delays[0]["A"] )
        m.connect( delays[-1]["B"], tank1["input"] )
        m.connect( boiler1["output"], gearvalve1["A"] )
        m.connect( gearvalve1["escape"], m["A*B"] )
        m.connect( tank1["output"], gearvalve1["gear"])
        
        m.connect( boiler2["output"], delays3[0]["A"])
        m.connect( delays3[-1]["B"], gearvalve4["A"] )
        m.connect( gearvalve4["gear"], manualgear["gear"] )
        m.connect( gearvalve4["B"], tank2["input"])
        manualgear.setValue(999)
        
        m.connect( m["A"],  gearvalve2["A"])
        m.connect( gearvalve2["gear"], tank2["output"])
        m.connect( gearvalve2["B"], pressuregear["input"] )
        
        m.connect( boiler3["output"], delays2[0]["A"])
        m.connect( delays2[-1]["B"],  gearvalve3["A"] )
        m.connect( gearvalve3["gear"], pressuregear["gear"] )
        m.connect( gearvalve3["B"], m["done"])
        
        
            
            
    def testSolution(self):
        result = False
        #self.m.dump()
        m = self.m
        for i in range(10000):
            #print (("*"*80)+"\n")*3
            #raw_input("next>")
            self.m.step()
            if self.m["done"].in_value == 1000:
                result = self.m.validate()
                break
        print m["A"].out_value, "*", m["B"].out_value, "=", m["A*B"].in_value   
        self.assert_(result)
        
        
        
DEBUG = 1
if __name__ == "__main__":
    if not DEBUG:
        unittest.main()
    else:
        suite = unittest.makeSuite(MultiplicationMachine)
        unittest.TextTestRunner(verbosity=2).run(suite)

        
    
  
